#!/usr/bin/python
# -*- coding: utf-8 -*-


from collections import deque
import select, socket, os, stat
import sys, tempfile, shutil, time
import re, errno
from syslog import *

__all__ = ['SCANNEDONLY', 'ScannedOnly', 'Options']

class SCANNEDONLY:
    CLEAN	= 0
    INFECTED= 1
    FAILED	= 2

class ScannedOnly(object):
    def __init__(self, opts):
        self.opts = opts
        self.__syslog()
        self.conn = self.__open()
        if opts.exclude:
            try:
                self.exclude = re.compile(opts.exclude)
            except re.error, reason:
                syslog(LOG_CRIT, "Warning: exclude pattern '%s' is not "
                                 "a valid regular expression pattern!: %s" %
                                 (opts.exclude, str(reason)))

    def __syslog(self):
        openlog(self.opts.prog, LOG_PID, LOG_DAEMON)
        if 0 <= self.opts.loglevel <= 3:
            logmask = (LOG_MASK(LOG_EMERG)| LOG_MASK(LOG_ALERT)| LOG_MASK(LOG_CRIT)|
                       LOG_MASK(LOG_ERR)| LOG_MASK(LOG_WARNING))
            if self.opts.loglevel > 0:
                logmask |= LOG_MASK(LOG_NOTICE)
            if self.opts.loglevel > 1:
                logmask |= LOG_MASK(LOG_INFO)
            if self.opts.loglevel > 2:
                logmask |= LOG_MASK(LOG_DEBUG)
        else:
            logmask = 0 # 0 means that the current logmask is not modified
        setlogmask(logmask)

    def __open(self):
        sockname = self.opts.portnum or self.opts.socket
        if isinstance(sockname, int):
            syslog(LOG_NOTICE, "Starting with udp port %d and %d threads" %
                               (sockname, self.opts.maxthreads))
            try:
                conn = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            except socket.error, reason:
                syslog(LOG_CRIT, "ABORT, %s" % str(reason))
                sys.exit(1)
            try:
                conn.bind(('', sockname))
            except socket.error, reason:
                syslog(LOG_CRIT, "ABORT, could not bind port %s: %s" %
                       (sockname, str(reason)))
                sys.exit(1)
        else:
            syslog(LOG_NOTICE, "Starting with socket %s and %d threads" %
                   (sockname, self.opts.maxthreads))
            open("/proc/sys/net/unix/max_dgram_qlen", 'w').write("200")
            try:
                conn = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
            except socket.error, reason:
                syslog(LOG_CRIT, "ABORT, %s" % str(reason))
                sys.exit(1)
            if os.path.exists(sockname):
                os.unlink(sockname)
            try:
                conn.bind(sockname)
            except socket.error, reason:
                syslog(LOG_CRIT, "ABORT, could not bind socket %s: %s" %
                       (sockname, str(reason)))
                sys.exit(1)
            os.chmod(sockname, 0666)
        sndsize = 524288
        try:
            conn.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, sndsize)
        except socket.error, reason:
            syslog(LOG_NOTICE, "Failed to increase socket buffer size to %d bytes: %s" %
                   (sndsize, str(reason)))
        else:
            newsize = conn.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF)
            if newsize != sndsize:
                syslog(LOG_NOTICE, "Failed to increase socket buffer size. "
                                   "Requested %d bytes, got %d bytes" % 
                                   (sndsize,newsize))
        return conn.makefile('r', 1)

    def __call__(self, scand):
        scand.setup()
        producers = dict((scanner.write_fd(), scanner) for scanner in scand.scanners())
        scanners = deque()
        consumers = dict()
        queue = deque()
        scanning = set()
        scanning_large = 0
        sizes = {}
        bigfilesize = self.opts.big * 1048576 # MB
        timeout = None

        while True:
            input, output, _ = select.select(consumers.keys() + [self.conn],
                                             producers.keys(), [], timeout)

            for conn in output:
                scanners.append(producers.pop(conn))

            for conn in input:
                if conn is self.conn:
                    path = self.read()
                    if not path in queue:
                        size = self.scansize(path)
                        if size:
                            if len(queue) > self.opts.queuelen:
                                syslog(LOG_INFO, "Warning: Received %s but skipped request; "
                                        "queue limit %d" % (name, self.opts.queuelen))
                                continue
                            queue.append(path)
                            syslog(LOG_INFO, "Received %s for scanning, %d files on queue" %
                                    (path, len(queue)))
                            sizes[path] = size
                else:
                    scanner, state = consumers.pop(conn)
                    try:
                        result, path, descr = scanner.consume(state)
                    except (TypeError, ValueError):
                        # request was not processed!
                        consumers[conn] = scanner, state
                    else:
                        self.handle(result, path, descr)
                        scanning.remove(path)
                        queue.remove(path)
                        if sizes.pop(path) > bigfilesize:
                            scanning_large -= 1
                        producers[scanner.write_fd()] = scanner

            while scanners:
                scanner = scanners.popleft()
                for path in queue:
                    if path in scanning:
                        continue
                    size = sizes[path]
                    if size > bigfilesize:
                        if scanning_large >= self.opts.maxlargethreads:
                            syslog(LOG_INFO, "large queue >= %d: "
                                             "postponing request to scan %s" %
                                             (self.opts.maxlargethreads, path))
                            continue
                        scanning_large += 1
                        syslog(LOG_DEBUG, "scanning large %s, queue length %d" %
                                (path, len(queue)))
                    else:
                        syslog(LOG_DEBUG, "scanning small %s, queue length %d" %
                                (path, len(queue)))
                    scanning.add(path)
                    # bind scanner to state
                    state = scanner.produce(path)
                    consumers[scanner.read_fd()] = scanner, state
                    break # continue withe the next scanner
                else:
                    # out of names, i.e. nothing to do
                    scanners.appendleft(scanner)
                    break

    def read(self):
        name = self.conn.readline()[:-1]    # chomp '\n'
        return os.path.normpath(name)

    def scansize(self, name):
        # returns None or the size of the file to be scanned
        if not os.path.isfile(name):
            syslog(LOG_INFO, "Not scanning %s: not a file or it doesn't exist" % name)
            return
        if not os.path.commonprefix((self.opts.scanroot, name)) == self.opts.scanroot:
            syslog(LOG_NOTICE, "Ignored request to scan %s; it is outside the scanroot %s" %
                    (name, self.opts.scanroot))
            return
        if self.opts.exclude and self.exclude.match(name):
            syslog(LOG_INFO, "Ignored request to scan %s; it matches the excludepattern %s" %
                    (name, self.opts.exclude))

        name_stat = os.stat(name)
        if name_stat.st_size < self.opts.smallfilesize: # default 256
            # the smallest known virus is 260 bytes and a lot of files are
            # smaller than 256 bytes, so this should be a huge performance booster
            syslog(LOG_INFO, "Not scanning %s: filesize smaller than %d" %
                    (name, self.opts.smallfilesize))
            self.handle_clean(name, name_stat)
            return
        cname = _scanned(name)
        if os.path.exists(cname):
            if os.path.isfile(cname):
                cname_stat = os.stat(cname)
                if name_stat.st_mtime < cname_stat.st_mtime:
                    return
            else:
                # anomaly, unlink it
                try:
                    os.unlink(cname)
                except OSError:
                    pass
        # if it does not exist, perhaps the previous scan failed, check for that
        fname = _failed(name)
        if os.path.exists(fname):
            # there has been a previous scan that failed
            if os.path.isfile(fname):
                # it is a regular file and it exists, check the timestamps
                fname_stat = os.stat(fname)
                if name_stat.st_mtime < fname_stat.st_mtime:
                    # .failed: file is newest, rescanning is useless
                    if fname_stat.st_mtime + 60 * 5 < time.time():
                        # the file did not change the last 5 minutes,
                        # it's probably not actively written to
                        # so move the file
                        syslog(LOG_INFO, "Not scanning %s, last scan failed >5 minutes ago, "
                                         "creating a warning file" % name)
                        self.handle_warning(name, "failed_to_virus-scan_",
                                            ".failed:", name_stat)
                    else:
                        syslog(LOG_INFO, "Not scanning %s, previous scan failed" % name)
                    return
                syslog(LOG_INFO, "Will rescan %s, "
                                 "file changed after last failed attempt" % name)
            else:
                # anomaly, unlink it
                try:
                    os.unlink(fname)
                except OSError:
                    pass
        return name_stat.st_size

    def handle(self, result, name, descr):
        try:
            sbuf = os.stat(name)
        except OSError:
            return
        if result == SCANNEDONLY.CLEAN:
            syslog(LOG_INFO, "%s is clean" % name)
            self.handle_clean(name, sbuf)
        elif result == SCANNEDONLY.INFECTED:
            syslog(LOG_WARNING, "WARNING: %s contains virus %s!" % (name, descr))
            virusprefix = "VIRUS:%s:found_in_" % descr.replace('/', '_')
            self.handle_warning(name, virusprefix, ".virus:", sbuf)
        elif result == SCANNEDONLY.FAILED:
            syslog(LOG_CRIT, "WARNING: failed to scan %s: %s" % (name, descr))
            self.handle_warning(name, "failed_to_virus-scan_", ".failed:", sbuf)
        else:
            raise RuntimeError, "Ooops"

    def handle_clean(self, name, sbuf):
        _touch(_scanned(name), sbuf)

    def handle_warning(self, name, warnprefix, moveprefix, sbuf):
        if self.opts.quarantainedir:
            shortname = os.path.basename(name)
            # trick (hack?) to force numbers as tempfile name-sequence
            tempfile._name_sequence = (str(n) for n in xrange(999999999))
            fd, movename = tempfile.mkstemp(dir = self.opts.quarantainedir,
                                            prefix = shortname)
            os.close(fd)
        else:
            movename = _fixname(name, moveprefix)
        warn = _fixname(name, warnprefix, ".txt")
        _touch(warn, sbuf)
        self.handle_clean(warn, sbuf)
        try:
            syslog(LOG_NOTICE, "moving %s to %s" %(name, movename))
            os.rename(name, movename)
            return
        except OSError, reason:
            if reason.errno == errno.EXDEV:
                try:
                    shutil.copyfile(name, movename)
                    _chownmod(movename, sbuf)
                    os.unlink(name)
                    return
                except (IOError, OSError), reason:
                    pass
        # something went wrong
        syslog(LOG_NOTICE, "failed to rename %s to %s: %s" % (name, movename, str(reason)))

# helper functions
def _fixname(name, prefix, suffix = ''):
    dirname, filename = os.path.split(name)
    return os.path.join(dirname, prefix + filename + suffix)

def _scanned(name):
    return _fixname(name, '.scanned:')

def _failed(name):
    return _fixname(name, '.failed:')

def _chownmod(name, sbuf):
    os.chown(name, sbuf.st_uid, sbuf.st_gid)
    os.chmod(name, stat.S_IMODE(sbuf.st_mode))

def _touch(name, sbuf):
    open(name, 'w').close()
    try:
        _chownmod(name, sbuf)
        os.utime(name, None)
    except OSError:
        syslog(LOG_WARNING, "Failed to create %s" % name)

def Options(*options):
    from optparse import OptionParser

    parser = OptionParser()

    parser.add_option("-p","--portnum", type="int",
                      help="The UDP port to listen on"
                           "(but by default a domain socket is used)")
    parser.add_option("-s","--socket",
                      help="The unix domain socket name to listen on. "
                            "The directory needs to exist! "
                            "(default: %default)",
                      default="/var/lib/scannedonly/scan")
    parser.add_option("-t","--maxthreads",
                      help="The total number of scanning threads. "
                           "Maximum one thread (see option maxlargethreads) will be "
                           "scanning large files. Maximum one thread might be updating "
                           "the virus database. So use at least 3 threads if you want "
                           "to guarantee that there is always a thread for small files "
                           "available. (default: %default)",
                      default=4, type="int")
    parser.add_option("-g","--maxlargethreads",
                      help="The maximum number of scanning threads that may scan large "
                           "files. (default: %default)",
                      default=1, type="int")
    parser.add_option("-l","--loglevel",
                      help="The log level from 0 (little logging) to 3 (verbose logging) "
                      "(default: %default)", default=1, type="int")
    parser.add_option("-b","--big",
                      help="Which size (and higher) files, in Mb, are treated as big "
                      "(default: %default)",
                      default=5, type="int")
    parser.add_option("-S","--smallfilesize",
                      help="Which size (and lower) files, in Bytes, are treated as small "
                      "(default: %default)",
                      default=256, type="int")
    parser.add_option("-m","--time",
                      help="The interval for status messages in the logs "
                      "(default: %default)",
                      default=60, type="float")
    parser.add_option("-r","--scanroot",
                      help="The base directory for which requests for scanning are "
                           "accepted. Requests to scan files outside this directory "
                           "are ignored. Using this option is very much advised! "
                           "(default: %default)",
                     default="/")
    parser.add_option("-d","--quarantainedir",
                      help="By default a file is renamed into .virus:file. If you want "
                           "all viruses in a central location specify a quarantainedir. "
                           "If a virus is found it will be moved to this directory.")
    parser.add_option("-q","--queuelen",
                      help="The maximum request queue length. If the maximum number of "
                           "requests are on the queue, new requests are ignored. "
                           "(default: %default)",
                      default=500, type="int")
    parser.add_option("-i","--pidfile",
                      help="The location for a PID file")
    parser.add_option("-e","--exclude",
                      help="Exclude scanning of files that match this (regular expression) "
                           "pattern. For example '/home/[^/]+/Maildir/.*'. You might want "
                           "to use the option veto files in your Samba configuration to "
                           "exclude these files in samba as well, otherwise samba will "
                           "keep telling that these files need scanning.")

    for opt in options:
        parser.add_option(opt)

    opts, args = parser.parse_args()
    opts.prog = parser.get_prog_name()
    return opts, args

if __name__ == '__main__':
    import sys
    opts, args = Options()
    print opts, args

